#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastorm']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg', 'luckydonald-utils>=0.83', 'typeguard', 'pydantic']

setup(name='fastorm',
      version='0.0.8',
      description='FastORM framework, easy to learn, fast to code',
      author='luckydonald',
      author_email='fastorm+code@luckydonald.de',
      url='https://github.com/luckydonald/fastorm',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.10.0',
     )
